import { apply } from '../apply';
import { curry2 } from '../curry';
import { curryN } from '../curryN';
/**
 * A placeholder for `partial`.
 * @name __
 */
export var __ = { '@@placeholder': true };
var isPlaceholder = function (x) { return x['@@placeholder'] === true; };
/**
 * Allows partially applying a function
 * @name partial(f: Function, args: List<any>): PartiallyAppliedFunction
 */
export var partial = curry2(function (f, args) {
    var fnLength = f.length;
    var argsLength = args.length;
    if (fnLength === 0)
        return f;
    if (argsLength === 0)
        return curryN(fnLength, f);
    var placeholderAmount = args.filter(isPlaceholder).length;
    var expectedLength = Math.max(0, fnLength - argsLength) + placeholderAmount;
    function partiallyApplied() {
        var otherArgs = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            otherArgs[_i] = arguments[_i];
        }
        if (placeholderAmount === 0)
            return apply(args.concat(otherArgs), f);
        var combinedArgs = Array(fnLength);
        for (var i = 0; i < combinedArgs.length; ++i)
            combinedArgs[i] = isPlaceholder(args[i]) ? otherArgs.shift() : args[i];
        return apply(combinedArgs.concat(otherArgs), f);
    }
    return curryN(expectedLength, partiallyApplied);
});
//# sourceMappingURL=partial.js.map