"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Memoizes a function.
 * @name memoize<F extends Function>(f: F): F
 */
exports.memoize = function (f) {
    var cache = new Map();
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        var key = args.reduce(function (x, y) { return x + JSON.stringify(y); }, '');
        if (cache.has(key))
            return cache.get(key);
        var result = f.apply(this, args);
        if (typeof result === 'function')
            result = exports.memoize(result);
        cache.set(key, result);
        return result;
    };
};
//# sourceMappingURL=memoize.js.map