"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var apply_1 = require("../apply");
var curry_1 = require("../curry");
var curryN_1 = require("../curryN");
/**
 * A placeholder for `partial`.
 * @name __
 */
exports.__ = { '@@placeholder': true };
var isPlaceholder = function (x) { return x['@@placeholder'] === true; };
/**
 * Allows partially applying a function
 * @name partial(f: Function, args: List<any>): PartiallyAppliedFunction
 */
exports.partial = curry_1.curry2(function (f, args) {
    var fnLength = f.length;
    var argsLength = args.length;
    if (fnLength === 0)
        return f;
    if (argsLength === 0)
        return curryN_1.curryN(fnLength, f);
    var placeholderAmount = args.filter(isPlaceholder).length;
    var expectedLength = Math.max(0, fnLength - argsLength) + placeholderAmount;
    function partiallyApplied() {
        var otherArgs = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            otherArgs[_i] = arguments[_i];
        }
        if (placeholderAmount === 0)
            return apply_1.apply(args.concat(otherArgs), f);
        var combinedArgs = Array(fnLength);
        for (var i = 0; i < combinedArgs.length; ++i)
            combinedArgs[i] = isPlaceholder(args[i]) ? otherArgs.shift() : args[i];
        return apply_1.apply(combinedArgs.concat(otherArgs), f);
    }
    return curryN_1.curryN(expectedLength, partiallyApplied);
});
//# sourceMappingURL=partial.js.map